#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['wwl']

package_data = \
{'': ['*']}

install_requires = \
['Cython', 'POT', 'numpy', 'scikit-learn', 'python-igraph']

setup(name='wwl',
      version='0.1.2',
      description='Wasserstein Weisfeiler-Lehman Graph Kernels',
      author='Matteo Togninalli',
      author_email='matteo.togninalli@bsse.ethz.ch',
      url='https://github.com/BorgwardtLab/WWL',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      python_requires='>=3.6',
     )
